

// import tspielfeld.*;
import java.util.Random; // im IE funktioniert mehtode nextInt(i) nicht, aber nextInt
// import tspielfeld.tspielfeld.SpalteVollException;

/** implementiert eine Mehtode zur Ermittlung des besten Zuges
 * erweitert tspielfeld
 * @author Robert Arnold
 */

public class SchlauesSpielfeld
    extends tspielfeld
    implements Runnable {
  public class AbbruchException
      extends Exception {};
  private SchlauesSpielfeld probierfeld;
  private Random zufall = new Random();
  private static int bewertunggewonnen = 100;
  public int zugvorschlag;
  public int level;
  public int bewertung[] = new int[maxspieler + 1];
  public boolean berechnungabbrechen;
  public int debug;
  public int angriffslust;

  /**
   * Legt ein leeres Spielfeld an. Vor Spielstart sollte allerdings neuesspiel(tiefe, angriff) aufgerufen werden.
   */
  public SchlauesSpielfeld() {
    neuesspiel();
    this.berechnungabbrechen = false;
  }

  /**
   * Legt ein neues Spielfeld an.
   * @param tiefe Anzahl voraus zu berechnender Halbzge
   * @param angriff Zahl zwischen 0 und 4, die die die Angriffslust des Computers festlegt
   */
  public void neuesspiel(int tiefe, int angriff) {
    super.neuesspiel();
    level = tiefe;
    switch (angriff) {
      case 0:
        this.angriffslust = 1;
        break;
      case 1:
        this.angriffslust = bewertunggewonnen / 4;
        break;
      case 2:
        this.angriffslust = bewertunggewonnen / 4 * 2;
        break;
      case 3:
        this.angriffslust = bewertunggewonnen / 4 * 3;
        break;
      case 4:
        this.angriffslust = bewertunggewonnen - 1;
        break;
      default:
        this.angriffslust = bewertunggewonnen - 1;
    }
    this.berechnungabbrechen = false;
  }

  public SchlauesSpielfeld(SchlauesSpielfeld f) {
    kopiere(f);
  }

  public void kopiere(SchlauesSpielfeld f) {
    super.kopiere(f);
//    for (int i=1; i<=maxspieler; i++) {this.bewertung[i]=f.bewertung[i];  } /* berflssig wg drei Zeilen tiefer? */
    this.level = f.level;
    this.debug = f.debug;
    this.berechnungabbrechen = false;
    this.angriffslust = f.angriffslust;
    for (int i = 1; i <= maxspieler; i++) {
      this.bewertung[i] = 0;
    }
    ;
  }

  public void berechnungabbrechen() {
    berechnungabbrechen = true;
  }

  public void run() {
    zugvorschlag = -1;
    try {
      zugvorschlag = zugvorschlagen(); // maxspieler vorheriger spieler this.spieler, level, this
    }
    catch (AbbruchException e) {
      zugvorschlag = 0;
    }
  }

  /**
   * Gibt besten Zug zurck. Diese Routine fr externe Aufrufe verwenden. Sie startet die Rekursion (besterzug).
   *
   * @return Ergebnisse werden von besterzug() bernommen. 0: Kein Zugvorschlag, es hat bereits der letzte Spieler gewonnen. Wenn -1 zurckgegeben wird, liegt vermutlich ein Programmfehler vor (Problem mit den Vergleichsbewerten).
   * @throws AbbruchException Berechnung durch Aufruf von berechungabbrechen() durch einen andere Thread abgebrochen.
   */
  public int zugvorschlagen() throws AbbruchException {
    if (hatspgewonnen() > 0) {
      return 0;
    }
    probierfeld = new SchlauesSpielfeld(this);
    int erg = besterzug(level, Integer.MAX_VALUE, Integer.MIN_VALUE);
    return rang2nr(erg);
  }

  /**
   * Gibt den besten Zug zurck.
   *
   * @param tiefe Tiefe der Rekursion (Anzahl Hablzge), sonnvoller weise >=2, da sonst Gegenergewinne in 1 Vollzug nicht erkannt werden
   * @param vergleichsbewertungG Bewertung des Gegnerzuges bei bisher bekanntem besten eigenen Zug aus Sicht des aufrufenden. Bei externem Aufruf muss Integer.MAX_VALUE hier eingesetzt werden oder besser zugvorschlagen() verwendet werden.
   * @param vergleichsbewertungI Bewertung des besten bisher bekannten eigenen Zuges aus Sicht des Aufrufenden. Bei externem Aufruf muss Integer.MIN_VALUE hier eingesetzt werden oder besser zugvorschlagen() verwendet werden.
   * @return 1 bis 7 = bester Zug, 0 = Feld voll, -1 = Ermittlung abgebrochen, weil durch die Vergleichsbewertungen erkennbar ist, dass der Spieler, der den Zug vorher gemacht hat, bereits eine bessere Mglichkeit hat als er erhalten wrde, wenn er den Zug machen wrde, fr den gerade der Gegnerzug ermittelt wird.
   * @throws AbbruchException wird geworfen, wenn this.berechnungabbrechen durch einen anderen Tread auf true gesetzt wurde.
   */
  public int besterzug(int tiefe, int vergleichsbewertungG,
                       int vergleichsbewertungI) throws AbbruchException {
    /** @todo Perfomance kann verbessert werden, indem nicht einfach von links nach rechts
     * die Spalten durchgegangen werden, sondern in der Reihenfolge der vermuteten besten Zge
     * (z.B. Mitte zu erst)  Das lsst sich grundstzlich leicht realisieren durch ein Array,
     * das die bersetzung der Zugnummern vornimmt. */
    int bestestellung = Integer.MIN_VALUE;
    int besterzug = 0;
    int b;

    Thread.yield();
    if (berechnungabbrechen) {
      throw new AbbruchException();
    }

    int sp = probierfeld.spieler;
    int alteGegnerBewertung; // enthlt die bisher beste Bewertung -> Variable "bestestellung" eigentlich berflssig
    int alteEigeneBewertung; // enthlt die Bewertung des Gegners beim bisher besten Spielzug - wird bentigt, weil besterzug() probierfeld.bewertung[] berschreibt
    probierfeld.bewertung[sp] = Integer.MIN_VALUE; // enthlt die bisher beste Bewertung - Werte wichtig fr die bergabe in Rekursion, damit nicht abgebrochen wird, wenn noch keine Zug ausgefhrt wurde.
    probierfeld.bewertung[2 - (sp - 1)] = Integer.MAX_VALUE; // enthlt die Bewertung des Gegners beim bisher besten Spielzug

    if (tiefe == 1) {
      for (int x2 = 1; x2 <= maxX; x2++) {
        try {
          probierfeld.zugschnell(rang2nr(x2));
          b = auswerten(sp, probierfeld);
          //     if (debug==1) {System.out.println("voher tiefe: " + tiefe + ", Probvezug: " + x2 +
          //                        ", Spieler: " + sp + ", bester Zug: " + besterzug +
          //                        ", Wertung: " + b); }

          if (b > bestestellung) {
            bestestellung = b;
            besterzug = x2;
            probierfeld.bewertung[sp] = b;
            probierfeld.bewertung[2 -
                (sp - 1)] = auswerten(probierfeld.spieler, probierfeld);
          }
          else {
            if (b == bestestellung) {
              b = auswerten(probierfeld.spieler, probierfeld);
              if (b < probierfeld.bewertung[probierfeld.spieler] || // Gegner schlechter?
                  (b == probierfeld.bewertung[probierfeld.spieler] // Gegner gleich -> Zufall
                   )) { //             && (zufall.nextInt()&15) < 7  ) ) { //  !!! wieder rein machen, wenn fertig !!! - nein, verbraucht nur Zeit. Da nur die Bewertungen gebraucht werden, nicht die tatshcliche Stellung ist der Zug egal (Bewertungen sind ja gleich). Tief=1 wird nie von auen aufgerufen, nur durch Rekursion. zufall sollte evtl. gleich sein in Routine >1 - nein.
                besterzug = x2;
                probierfeld.bewertung[probierfeld.spieler] = b;
              }
            }
          }
          probierfeld.zugzurueck(rang2nr(x2));
          if ( (vergleichsbewertungG < b &&
                vergleichsbewertungI >= probierfeld.bewertung[2 - (sp - 1)]) // Aufrufender wrde seinen Gegner durch einen anderen vorhergehenden Zug schlechter stellen knnen und sich selbst gleich oder besser
              ||
              (vergleichsbewertungG <= b &&
               vergleichsbewertungI > probierfeld.bewertung[2 - (sp - 1)])) { // Aufrufender wrde sich durch einen anderen vorgehenden Zug besser stellen und den Gegner gleich oder schlechter
            besterzug = -1;
            break;
          }  // vorheriger (Gegner-) Zug kommt nicht in Frage, weil er sich dadurch schlechter stellen wrde als durch einen Zug, der bereits berechnet wurde. Darum brauchen die aktuellen Zugmglichkeiten nicht komplett durchgerechnet zu werden.
        }
        catch (SpalteVollException f) { /* Spalte voll */}
      }
    }

    if (tiefe > 1) {
      for (int x2 = 1; x2 <= maxX; x2++) {
        try {
          probierfeld.zugschnell(rang2nr(x2));
          //   b = auswerten(sp, probierfeld);  // hier gengt eigentlich hat gewonnen? - ja aber, dann wrde nicht mehr zum eigenen Vorteil gespielt, falls der andere die Mglichkeit bersieht, zu gewinnen.
//      int b2 = probierfeld.hatspgewonnen();
          alteGegnerBewertung = probierfeld.bewertung[2 - (sp - 1)];
          alteEigeneBewertung = probierfeld.bewertung[sp];
          if (probierfeld.hatspgewonnen() == 0) { // weder gewonnen noch unentschieden
            int bz = besterzug(tiefe - 1, alteGegnerBewertung,
                               alteEigeneBewertung); // Rekursion Achtung: ndert bewertung
            b = probierfeld.bewertung[sp];
            if (bz == 0) {
              b = auswerten(sp, probierfeld);
            } // spielfeld voll
            if (debug == 1) {
              System.out.println("voher tiefe: " + tiefe + ", Probvezug: " + x2 +
                                 ", Spieler: " + sp + ", bester Zug: " +
                                 besterzug +
                                 ", Wertung: " + b);
            }
          }
          else { // Gewonnen oder unentschieden
            // hier sollte evtl. noch Behandlung rein, was passieren soll, wenn unentschieden -> ist aber nicht unbedingt notwendig, da es beim nchsten Zug durch SpalteVollExeption ehe auffllt.
            probierfeld.bewertung[2 -
                (sp - 1)] = -bewertunggewonnen * bewertunggewonnen *
                bewertunggewonnen * tiefe;
            // muss fr alle anderen Spieler, die verloren haben auch gesetzt werden maxspieler
            b = bewertunggewonnen * bewertunggewonnen * bewertunggewonnen *
                tiefe; // nahes gewinnen besser als fernes
            probierfeld.bewertung[sp] = b;
          }
          if (b > bestestellung) {
            bestestellung = b;
            besterzug = x2;
          }
          else {
            if (b == bestestellung) {
              if (probierfeld.bewertung[2 - (sp - 1)] < alteGegnerBewertung || // Gegnger schlechter?
                  (probierfeld.bewertung[2 - (sp - 1)] == alteGegnerBewertung // Gegner gleich -> Zufall
                   && (zufall.nextInt() & 15) < 5)) { // zufall sollte evtl. gleich sein in Routine ==1
                besterzug = x2;
              }
              else { // stellung nicht besser, alte Bewertung speichern
                probierfeld.bewertung[sp] = alteEigeneBewertung;
                probierfeld.bewertung[2 - (sp - 1)] = alteGegnerBewertung;
              }
            }
            else {
              probierfeld.bewertung[sp] = alteEigeneBewertung;
              probierfeld.bewertung[2 - (sp - 1)] = alteGegnerBewertung;
            }
          }
          probierfeld.zugzurueck(rang2nr(x2));
          if ( (vergleichsbewertungG < b &&
                vergleichsbewertungI >= probierfeld.bewertung[2 - (sp - 1)]) // Aufrufender wrde seinen Gegner durch einen anderen vorhergehenden Zug schlechter stellen knnen und sich selbst gleich oder besser
              ||
              (vergleichsbewertungG <= b &&
               vergleichsbewertungI > probierfeld.bewertung[2 - (sp - 1)])) { // Aufrufender wrde sich durch einen anderen vorgehenden Zug besser stellen und den Gegner gleich oder schlechter
            besterzug = -1;
            break;
          }
        }
        catch (SpalteVollException f) { /* Spalte voll */}
      }
    }
    return besterzug;
  }

  static int heuristik[] = {
      0, 4, 3, 5, 2, 6, 1, 7, 8};
  private int rang2nr(int rangzug) {
    int zugecht = (rangzug > 0 && rangzug <= maxX ? heuristik[rangzug] :
                   rangzug);
    return zugecht;
  }

// Diese Routine kann verwendet werden, wenn auswerten verfeinert werden soll.
  private int auswerten(int farbe, tspielfeld feld) {
    // int angriffslust = (bewertunggewonnen -1)*4/5; // hier *1 fr volle Angriffslus und =1 fr totale Defensive
    int kw = bewertunggewonnen * bewertunggewonnen +
        auswerten1(farbe, feld) * angriffslust -
        auswerten1(2 - (farbe - 1), feld) * (bewertunggewonnen - angriffslust);
    //      ^^^^^^^^^^^^^^^^^^^^^^ damit immer positiv kw >0 && kw < bewertunggewonnen * bewertunggewonnen + bewertunggewonnen * angriffslust + bewertunggewonnen * 2 [verlieren = -2* bewertunggewonnen]* (bewertunggewonnen - angriffslust)
    return kw;
  }

  private int auswerten1(int farbe, tspielfeld feld) {
    int x = 0;
    int y = 0;
    int erg = 0;

    for (x = 1; x <= maxX; x++) {
      for (y = 1; y <= maxY; y++) {
        if (feld.daten[x][y] == 0) {
          break; // alle leeren Felder bewerten?
        }
        erg = erg + einbewerten(x, y, farbe, feld);
      }
    }
    if (erg == 0) {
      erg = -10;
    } // weder gewonnen, verloren, noch sonst eine
    // Mglichkeit von den bestehenden Steinen aus 4 in eine Reihe zu bekommen
    // bedenke: eigentlich sollte das ziemlich schlecht sein
    //          aber wenn das Spielfeld voll ist
    //          ist es besser als verloren
    // solte aber nicht negativ sein, weil besterzug sonst "verloren" annimmt -> nachdenken

    return erg;
  }

  /**
   * Ermittelt, ob durch den letzten Zug gewonnen wurde
   *
   * @return >0: Anzahl von 4 in einer Reihe in der Farbe des letzten Spielers, 0: Spieler mit der Farbe des letzten Zuges hat nicht gewonnen, -1: unentschieden = Spielfeldvoll
   */
  public int hatspgewonnen() {
    /** @todo Performance kann leicht verbessert werden, indem nur der letzte Zug ausgewertet wird. - Evtl. eine Variante hatspgewonnenschnell */
//  int w=auswerten(sp, this);
//  if (w >= 500) {return 1;}
//  if (w <= -10000) {return -1;} achtung: neu: -1=unentschieden.
    int x = 0;
    int y = 0;
    int erg = 0;
    int belegt = 0;

    for (x = 1; x <= maxX; x++) {
      for (y = 1; y <= maxY; y++) {
        if (daten[x][y] == 0) {
          break; // alle leeren Felder bewerten?
        }
        erg = erg + eingewonnen(x, y); // evtl. abbrechen, wenn 1mal gewonnen
        belegt++;
      }
    }
    if (belegt >= maxX * maxY && erg == 0) {
      erg = -1;
    } // spielfeldvoll, niemand gewonnen -> unentschieden.
    return erg;
  }

  private int eingewonnen(int x, int y) {
    // Achtung: Routine bemerkt nur Gewonnene in vom Anfangsstein in 1 Richtung! - Wenn alle Steine Untersucht werden -> kein Problem, wenn aber nur neu gelegter Stein untersucht wird, muss das gendert werden!!!
    int farbe = 2 - (this.spieler - 1); //maxspieler
    if (daten[x][y] != farbe) {
      return 0;
    }
    boolean[] Gewo = new boolean[9];
    for (int i = 1; i <= 8; i++) {
      Gewo[i] = true;
    }
    int anzahlG = 8;
    for (int i = 1; i <= 3; i++) {
      if (Gewo[1] && daten[x + i][y] != farbe) {
        Gewo[1] = false;
        anzahlG--;
      } //rechts
      if (Gewo[2] && daten[x - i][y] != farbe) {
        Gewo[2] = false;
        anzahlG--;
      } //links
      if (Gewo[3] && daten[x][y + i] != farbe) {
        Gewo[3] = false;
        anzahlG--;
      } //oben
      if (Gewo[4] && daten[x][y - i] != farbe) {
        Gewo[4] = false;
        anzahlG--;
      } //unten
//diagonal
      if (Gewo[5] && daten[x + i][y + i] != farbe) {
        Gewo[5] = false;
        anzahlG--;
      } //rechts oben
      if (Gewo[6] && daten[x - i][y - i] != farbe) {
        Gewo[6] = false;
        anzahlG--;
      } //links unten
      if (Gewo[7] && daten[x - i][y + i] != farbe) {
        Gewo[7] = false;
        anzahlG--;
      } //links oben
      if (Gewo[8] && daten[x + i][y - i] != farbe) {
        Gewo[8] = false;
        anzahlG--;
      } //rechts unten
    }
    return anzahlG * bewertunggewonnen;
  }

  private int einbewerten(int x, int y, int farbe, tspielfeld feld) {
    /** @todo performance verbessern */
//  int erg;
//  int gewonnen = countpossibilities(x,y,farbe,farbe, feld);
//  int verloren = countpossibilities(x,y,2-(farbe-1),2-(farbe-1), feld);
//  int wertung2  = countpossibilities(x,y,farbe,0, feld);
// if (gewonnen > 0)
//   {erg=500;}
// if (verloren > 0)
//   {erg=-5000;} // bedenke: weil in 1 Viererreihe immer von 2 Steinen aus
    // 4 in einer Reihe sind, ist die Bewertung eines Gewinns 2*5000.
    int wertung = countpossibilitiesschnell(x, y, farbe, feld);
//  int wertung3  = countpossibilitiesschnell(x,y,farbe, feld);
    return wertung;
  }

  private int countpossibilities(int x, int y, int farbe, int farbe2,
                                 tspielfeld feld) {
    int erg = 0;
    int x2 = 0;
    int y2 = 0;
    int i;
    int anzahl = 0;
    if (feld.daten[x][y] != farbe && feld.daten[x][y] != farbe2) {
      return 0;
    }
    for (y2 = y; y2 >= max(1, y - 3); y2--) { // 4 nach unten mglich?
      if (feld.daten[x][y2] != farbe && feld.daten[x][y2] != farbe2) {
        break;
      }
    }
    if (y >= y2 + 4) {
      anzahl++;
    }
    for (y2 = y; y2 <= min(maxY, y + 3); y2++) { // 4 nach oben mglich?
      if (feld.daten[x][y2] != farbe && feld.daten[x][y2] != farbe2) {
        break;
      }
    }
    if (y <= y2 - 4) {
      anzahl++;
    }
    for (x2 = x; x2 >= max(1, x - 3); x2--) { // 4 nach links mglich?
      if (feld.daten[x2][y] != farbe && feld.daten[x2][y] != farbe2) {
        break;
      }
    }
    if (x >= x2 + 4) {
      anzahl++;
    }
    for (x2 = x; x2 <= min(maxX, x + 3); x2++) { // 4 nach rechts mglich?
      if (feld.daten[x2][y] != farbe && feld.daten[x2][y] != farbe2) {
        break;
      }
    }
    if (x <= x2 - 4) {
      anzahl++;
    }
// diagonal
    for (i = 0; i <= 3; i++) { // 4 nach links oben mglich?
      if (feld.daten[x - i][y + i] != farbe &&
          feld.daten[x - i][y + i] != farbe2) {
        break;
      }
    }
    if (i >= 4) {
      anzahl++;
    }
    for (i = 0; i <= 3; i++) { // 4 nach links unten mglich?
      if (feld.daten[x - i][y - i] != farbe &&
          feld.daten[x - i][y - i] != farbe2) {
        break;
      }
    }
    if (i >= 4) {
      anzahl++;
    }
    for (i = 0; i <= 3; i++) { // 4 nach rechts unten mglich?
      if (feld.daten[x + i][y - i] != farbe &&
          feld.daten[x + i][y - i] != farbe2) {
        break;
      }
    }
    if (i >= 4) {
      anzahl++;
    }
    for (i = 0; i <= 3; i++) { // 4 nach rechts oben mglich?
      if (feld.daten[x + i][y + i] != farbe &&
          feld.daten[x + i][y + i] != farbe2) {
        break;
      }
    }
    if (i >= 4) {
      anzahl++;
    }

    erg = anzahl;
    return erg;
  }

  private int min(int x, int y) {
    return x < y ? x : y;
  }

  private int max(int x, int y) {
    return x > y ? x : y;
  }

  private int countpossibilitiesschnell(int x, int y, int wertungsfarbe,
                                        tspielfeld feld) {
    if (feld.daten[x][y] == 0) {
      return 0;
    }
    int farbe = feld.daten[x][y];
    boolean[] Frei = new boolean[9];
    boolean[] Gewo = new boolean[9];
    for (int i = 1; i <= 8; i++) {
      Frei[i] = true;
      Gewo[i] = true;
    }
    int anzahlG = 8;
    int anzahlF = 8;
    for (int i = 1; i <= 3; i++) {
      if (Gewo[1] && feld.daten[x + i][y] != farbe) {
        Gewo[1] = false;
        anzahlG--;
      } //rechts
      if (Frei[1] && feld.daten[x + i][y] != 0 && feld.daten[x + i][y] != farbe) {
        Frei[1] = false;
        anzahlF--;
      }
      if (Gewo[2] && feld.daten[x - i][y] != farbe) {
        Gewo[2] = false;
        anzahlG--;
      } //links
      if (Frei[2] && feld.daten[x - i][y] != 0 && feld.daten[x - i][y] != farbe) {
        Frei[2] = false;
        anzahlF--;
      }
      if (Gewo[3] && feld.daten[x][y + i] != farbe) {
        Gewo[3] = false;
        anzahlG--;
      } //oben
      if (Frei[3] && feld.daten[x][y + i] != 0 && feld.daten[x][y + i] != farbe) {
        Frei[3] = false;
        anzahlF--;
      }
      if (Gewo[4] && feld.daten[x][y - i] != farbe) {
        Gewo[4] = false;
        anzahlG--;
      } //unten
      if (Frei[4] && feld.daten[x][y - i] != 0 && feld.daten[x][y - i] != farbe) {
        Frei[4] = false;
        anzahlF--;
      }
//diagonal
      if (Gewo[5] && feld.daten[x + i][y + i] != farbe) {
        Gewo[5] = false;
        anzahlG--;
      } //rechts oben
      if (Frei[5] && feld.daten[x + i][y + i] != 0 &&
          feld.daten[x + i][y + i] != farbe) {
        Frei[5] = false;
        anzahlF--;
      }
      if (Gewo[6] && feld.daten[x - i][y - i] != farbe) {
        Gewo[6] = false;
        anzahlG--;
      } //links unten
      if (Frei[6] && feld.daten[x - i][y - i] != 0 &&
          feld.daten[x - i][y - i] != farbe) {
        Frei[6] = false;
        anzahlF--;
      }
      if (Gewo[7] && feld.daten[x - i][y + i] != farbe) {
        Gewo[7] = false;
        anzahlG--;
      } //links oben
      if (Frei[7] && feld.daten[x - i][y + i] != 0 &&
          feld.daten[x - i][y + i] != farbe) {
        Frei[7] = false;
        anzahlF--;
      }
      if (Gewo[8] && feld.daten[x + i][y - i] != farbe) {
        Gewo[8] = false;
        anzahlG--;
      } //rechts unten
      if (Frei[8] && feld.daten[x + i][y - i] != farbe &&
          feld.daten[x + i][y - i] != 0) {
        Frei[8] = false;
        anzahlF--;
      }
    }
    if (farbe != wertungsfarbe) {
      anzahlG *= -2;
      anzahlF = 0;
    } // wenn fr gegenfarbe gewertet werden sollte, Gewinnbewertung negativ und verdoppelt.
    // fr Gegenger wird nur gewonnen untersucht -> Performance verbessern durch eigene Routine fr Gewonnen?
    return anzahlF + anzahlG * bewertunggewonnen;
  }

}
